from __future__ import annotations

from enum import Enum
from types import MethodType
from typing import Callable

import gc

import asyncio
import weakref


__all__ = (
    "DispatchFramework",
    "listen",
    "ListenerClass",
    "WaitForCheck",
)


# T = TypeVar("T")


# _logger = logging.getLogger(__name__)


# class ClassMaker:
#     """
#     Used to delay creating an object from a given class with the given args, kwargs, and optionally the
#     parent class object.
#
#     Can only be used inside a class subclassing :class:`ClassMakerMixin`
#
#     Parameters
#     ----------
#     cls: type
#         Class to instantiate and return when called.
#     args: tuple[Any, ...]
#         Positional arguments to instantiate the class with.
#     kwargs: dict[str, Any]
#         Keyword arguments to instantiate the class with.
#     """
#     name: str | None
#     cls: type
#     args: tuple[Any, ...]
#     kwargs: dict[str, Any]
#
#     def __init__(self, cls: type, *args, **kwargs):
#         self.cls = cls
#         self.args = args
#         self.kwargs = kwargs
#
#     def __set_name__(self, owner: type[ClassMakerMixin], name: str):
#         if issubclass(owner, ClassMakerMixin):
#             self.name = name
#             owner._added_makers[name] = self
#         else:
#             raise ValueError("A class utilizing ClassMaker must subclass ClassMakerMixin")
#
#     def __call__(self, obj: ClassMakerMixin | None = None):
#         if obj:
#             return self.cls(obj, *self.args, **self.kwargs)
#         else:
#             return self.cls(*self.args, **self.kwargs)


# class ClassMakerMixin:
#     """
#     Subclassed to automatically have added :class:`ClassMaker` objects assigned in the class to fully instantiate
#     themselves as per-instance objects.
#     """
#     _added_makers: dict[str, ClassMaker] = {}
#
#     def __new__(cls):
#         ret = super(ClassMakerMixin, cls).__new__(cls)
#
#         for name, maker in cls._added_makers.items():
#             ret.__dict__[name] = maker(ret)
#
#         return ret


# class IterQueue(queue.Queue, Generic[T]):
#     def __iter__(self):
#         return self
#
#     def __next__(self) -> T:
#         try:
#             return self.get_nowait()
#         except queue.Empty:
#             raise StopIteration


EVENT_TYPE = str | type | Enum


# @ClassMaker
class ListenerClass:
    event: str | type
    callback: Callable

    def __init__(self, event: str | type, callback: Callable):
        self.event = event
        self.callback = callback


class WaitForCheck:
    def __init__(self, future: asyncio.Future, check: Callable[..., bool]):
        self.future: asyncio.Future = future
        self.check: Callable[..., bool] = check

    def __await__(self):
        yield from self.future


class DispatchFramework:
    __permanent_listeners__: dict[str | type, set[Callable]]
    __temporary_listeners__: dict[str | type, set[WaitForCheck]]
    # __temporary_listeners__: dict[str | type, weakref.WeakSet[WaitForCheck]]

    def __new__(cls, *args, **kwargs):
        new_cls = super(DispatchFramework, cls).__new__(cls)
        new_cls._discover_listeners()
        return new_cls

    def _discover_listeners(self):
        self.__permanent_listeners__ = {}
        self.__temporary_listeners__ = {}
        for base in reversed(self.__class__.__mro__):
            for elem, value in base.__dict__.items():
                if isinstance(value, staticmethod):
                    value = value.__func__

                if isinstance(value, ListenerClass):
                    if not self.__permanent_listeners__.get(value.event):
                        self.__permanent_listeners__[value.event] = set()

                    base.__dict__[elem] = MethodType(value.event, self)
                    self.__permanent_listeners__[value.event].add(base.__dict__[elem])

    def add_listener(self, func: ListenerClass | Callable, event: EVENT_TYPE):
        if isinstance(func, ListenerClass):
            event = event or func.event
            func = func.callback

        if event not in self.__permanent_listeners__:
            self.__permanent_listeners__[event] = set()

        self.__permanent_listeners__[event].add(func)

    def remove_listener(self, func: ListenerClass | Callable, event: str | type | Enum) -> bool:
        if isinstance(func, ListenerClass):
            event = event or func.event
            func = func.callback

        if func in self.__permanent_listeners__.get(event, set()):
            self.__permanent_listeners__[event].remove(func)
            return True
        else:
            return False

    def listen(self, event: str | type | Enum | None = None):
        def wrapper(func: Callable):
            name = event or func.__name__
            self.add_listener(func=func, event=name)
            return func

        return wrapper

    async def wait_for(self, event: type | str | Enum, check: Callable[..., bool] | None = None, timeout: float | None = None):
        future = asyncio.get_running_loop().create_future()
        if not check:
            def _check(*args, **kwargs):
                return True

            check = _check

        if event not in self.__temporary_listeners__:
            self.__temporary_listeners__[event] = set()

        check = WaitForCheck(future, check)
        self.__temporary_listeners__[event].add(check)

        try:
            ret = await asyncio.wait_for(future, timeout=timeout)
        except Exception as e:
            raise e
        else:
            return ret
        finally:
            self.__temporary_listeners__[event].discard(check)

    def dispatch(self, event: type | str | Enum, *args, **kwargs):
        loop = asyncio.get_running_loop()
        if event in self.__temporary_listeners__:
            temp_listeners = self.__temporary_listeners__[event].copy()
            for listener in temp_listeners:
                if listener.future.cancelled():
                    self.__temporary_listeners__[event].remove(listener)
                else:
                    try:
                        result = listener.check(*args, **kwargs)
                    except Exception as e:
                        listener.future.set_exception(e)
                    else:
                        if result:
                            match len(args):
                                case 0:
                                    listener.future.set_result(None)
                                case 1:
                                    listener.future.set_result(args[0])
                                case _:
                                    listener.future.set_result(args)

                            self.__temporary_listeners__[event].remove(listener)

        for listener in self.__permanent_listeners__.get(event, set()):
            loop.create_task(listener(*args, **kwargs))


def listen(listen_for: str | type | Enum):
    def wrapper(func: Callable):
        if asyncio.iscoroutinefunction(func):
            ret = ListenerClass(listen_for, func)
            ret.__call__ = func
            return ret
        else:
            raise ValueError("Given function is not a coroutine.")
    return wrapper




from enum import Enum, unique


__all__ = (
    "CoreEvents",
    "EngineEvents",
    "IPCClassType",
    "IPCPayloadType",
)


# @unique
class CoreEvents(Enum):
    ROUTED_CONN_INCOMING = "ROUTED_CONN_INCOMING"
    ROUTED_CONN_CONNECTION = "ROUTED_CONN_CONNECTION"
    ROUTED_CONN_MESSAGE = "ROUTED_CONN_MESSAGE"


class EngineEvents(Enum):
    # Base Engine Events
    ENGINE_READY = "ENGINE_READY"
    ENGINE_CLOSING = "ENGINE_CLOSING"
    # Node addition/removal.
    WS_NODE_ADDED = "WS_NODE_ADDED"
    NODE_ADDED = "NODE_ADDED"
    WS_NODE_REMOVED = "WS_NODE_REMOVED"
    NODE_REMOVED = "NODE_REMOVED"
    # Role addition/removal.
    EXTERNAL_ROLE_ADDED = "EXTERNAL_ROLE_ADDED"
    LOCAL_ROLE_ADDED = "LOCAL_ROLE_ADDED"
    ROLE_ADDED = "ROLE_ADDED"
    EXTERNAL_ROLE_REMOVED = "EXTERNAL_ROLE_REMOVED"
    LOCAL_ROLE_REMOVED = "LOCAL_ROLE_REMOVED"
    ROLE_REMOVED = "ROLE_REMOVED"
    # Device addition/removal.
    EXTERNAL_DEVICE_ADDED = "EXTERNAL_DEVICE_ADDED"
    LOCAL_DEVICE_ADDED = "LOCAL_DEVICE_ADDED"
    DEVICE_ADDED = "DEVICE_ADDED"
    EXTERNAL_DEVICE_REMOVED = "EXTERNAL_DEVICE_REMOVED"
    LOCAL_DEVICE_REMOVED = "LOCAL_DEVICE_REMOVED"
    DEVICE_REMOVED = "DEVICE_REMOVED"
    # Packet handling.
    WS_PACKET = "WS_PACKET"
    LOCAL_PACKET = "LOCAL_PACKET"
    PACKET = "PACKET"


@unique
class IPCClassType(Enum):
    ENGINE = 0
    ROLE = 1
    DEVICE = 2


@unique
class IPCPayloadType(Enum):
    COMMUNICATION = 0
    """Valid destination: ENGINE, ROLE, DEVICE
    
    Packet data: IPCPacket dictionary
    """
    DISCOVERY = 1
    """Valid destination: ENGINE

    Packet data: "Node UUID"
    """
    ROLE_ADD = 2
    """Valid destination: ENGINE

    Packet data: "Role name here"
    """
    ROLE_REMOVE = 3
    """Valid destination: ENGINE

    Packet data: "Role name here"
    """
    DEVICE_ADD = 4
    """Valid destination: ENGINE

    Packet data: {"uuid": "Device UUID Here", "role": "Role name here"}
    """
    DEVICE_REMOVE = 5
    """Valid destination: ENGINE

    Packet data: {"uuid": "Device UUID Here", "role": "Role name here"}
    """
    COMMUNICATION_REQUEST = 6
    """Valid destination: ENGINE, ROLE, DEVICE
    
    Packet data: None
    """
    COMMUNICATION_ACCEPTED = 7
    """Valid destination: ENGINE, ROLE, DEVICE (specifically in response to a COMMUNICATION_REQUEST)
    
    Packet data: UUID of the REQUESTing IPCConnection
    """
    COMMUNICATION_DENIED = 8
    """Valid destination: ENGINE, ROLE, DEVICE (specifically in response to a COMMUNICATION_REQUEST)

    Packet data: UUID of the REQUESTing IPCConnection
    """
    COMMUNICATION_REDIRECT = 9
    """Valid destination: ENGINE, ROLE, DEVICE (specifically, origin_type of the IPCConnection to redirect)
    
    Packet data: {
        "destination_type": IPCClassType.value, 
        "destination_name": "UUID/name here", 
        "destination_node": "Node UUID"
    }
    """



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

import uuid

from . import connection


if TYPE_CHECKING:
    from .engine import IPCEngine
    from .role import Role


__all__ = (
    "Device",
)


logger = getLogger(__name__)


class Device(connection.IPCCore):
    def __init__(self, uuid_override: str | None = None):
        super().__init__()
        self._uuid: str = uuid_override or uuid.uuid1().hex
        self._engine: IPCEngine | None = None
        """Set when the Device is added to an Engine."""
        self._role: Role | None = None
        """Set when the Device is added to a Role."""

    @property
    def role(self) -> Role | None:
        return self._role

    def set_role(self, role: Role):
        self._role = role

    @property
    def engine(self) -> IPCEngine | None:
        return self._engine

    def set_engine(self, engine: IPCEngine):
        self._engine = engine
        self.on_set_engine(engine)

    def on_set_engine(self, engine: IPCEngine):
        """Ran when the engine is set. Exists for the sole purpose of being optionally overridden."""
        pass

    @property
    def uuid(self) -> str:
        return self._uuid



from __future__ import annotations

import aiohttp
import asyncio
import signal
import uuid

from aiohttp import web
from logging import getLogger
from typing import Coroutine, TYPE_CHECKING

from .connection import IPCRawConnection, IPCRoutedConnection, IPCCore, IPCPacket, get_requestor_info
from .enums import CoreEvents, EngineEvents, IPCClassType, IPCPayloadType


if TYPE_CHECKING:
    from .device import Device
    from .role import Role


__all__ = (
    "ConnectionMap",
    "IPCEngine",
)


logger = getLogger(__name__)


CONNECT_RETRY_SLEEP = 10  # Time in seconds.
ENGINE_IPC_ROUTE = "engine/ipc"
ON_SIGNAL_CLOSE = (signal.SIGINT, signal.SIGTERM)
"""On what signals should the engine attempt to close itself with."""


class ConnectionMap:

    class MISSING:
        pass

    def __init__(self, engine: IPCEngine):
        self._engine = engine

        self._remote_nodes: dict[str, web.WebSocketResponse | aiohttp.ClientWebSocketResponse] = {}
        """{"Node UUID": (Client)WebSocketResponse object}"""
        self._remote_roles: dict[str, set[str]] = {}
        """{"Role Name": {"Node UUID 1", "Node UUID 2", "etc."}}"""
        self._remote_devices: dict[str, tuple[str, str]] = {}
        """{"Device UUID": ("Role Name", "Node UUID")}"""

    @property
    def remote_nodes(self) -> dict[str, web.WebSocketResponse | aiohttp.ClientWebSocketResponse]:
        return self._remote_nodes.copy()

    @property
    def remote_roles(self) -> dict[str, set[str]]:
        return self._remote_roles.copy()

    @property
    def remote_devices(self) -> dict[str, tuple[str, str]]:
        return self._remote_devices.copy()

    @property
    def local_roles(self) -> set[str]:
        # return self._local_roles.copy()
        return set(self._engine._roles.keys())

    @property
    def local_devices(self) -> set:
        ret = set()
        for role in self._engine._roles.values():
            for device in role.local_devices.values():
                ret.add(device.uuid)

        return ret

    @property
    def nodes(self) -> list[str]:
        return list(self.remote_nodes.keys()) + [self._engine.uuid, ]

    @property
    def roles(self) -> list[str]:
        return list(self.local_roles) + list(self._remote_roles.keys())

    @property
    def devices(self) -> list[str]:
        return list(self.local_devices) + list(self._remote_devices.keys())

    @property
    def full_device_info(self) -> dict[str, tuple[str, str]]:
        ret = self._remote_devices.copy()
        ret.update({
            device_uuid: (device.role.name, self._engine.uuid) for device_uuid, device in self._engine.devices.items()
        })
        return ret

    def clean(self):
        for node_uuid, ws in self.remote_nodes.items():
            if ws.closed:
                self.remove_node(node_uuid)

    def add_node(self, ws: web.WebSocketResponse | aiohttp.ClientWebSocketResponse, node_uuid: str) -> bool:
        if node_uuid in self._remote_nodes:
            return False
        else:
            self._remote_nodes[node_uuid] = ws
            return True

    def remove_node(self, rm_node_uuid: str) -> bool:
        if rm_node_uuid not in self._remote_nodes:
            return False

        logger.debug("Fully removing node %s from NodeMap.", rm_node_uuid)

        # Removes the node UUID from the dict of remote roles.
        for role_name, node_list in self._remote_roles.items():
            if rm_node_uuid in node_list:
                node_list.remove(rm_node_uuid)
                self._engine.events.dispatch(EngineEvents.EXTERNAL_ROLE_REMOVED, role_name, rm_node_uuid)

        # Removes devices from the dict if the node they were attached to is no longer connected.
        for device_uuid, role_node in self.remote_devices.items():
            if role_node[1] == rm_node_uuid:
                logger.debug("Removing device %s", device_uuid)
                self._remote_devices.pop(device_uuid)
                self._engine.events.dispatch(EngineEvents.EXTERNAL_DEVICE_REMOVED, device_uuid, role_node)

        # Finally, removes the node UUID and websocket object for it from the dict.
        self._remote_nodes.pop(rm_node_uuid)
        self._engine.events.dispatch(EngineEvents.NODE_REMOVED, rm_node_uuid)

        return True

    def add_external_role(self, role_name: str, node_uuid: str):
        if node_uuid in self._remote_nodes:
            if role_name not in self._remote_roles:
                self._remote_roles[role_name] = set()

            self._remote_roles[role_name].add(node_uuid)
            logger.debug("Added Node %s to list for external role %s", node_uuid, role_name)
        else:
            raise ValueError(f"Node UUID {node_uuid} not found, cannot add it to the set for role {role_name}.")

    def add_external_device(self, device_uuid: str, device_role: str, node_uuid: str):
        if node_uuid in self._remote_nodes:
            self._remote_devices[device_uuid] = (device_role, node_uuid)
        else:
            raise ValueError(
                f"Node UUID {node_uuid} not found, cannot add device with UUID {device_uuid} and role {device_role}."
            )

    def resolve_node_conn(self, node_uuid: str) -> aiohttp.ClientWebSocketResponse | web.WebSocketResponse | None:
        if node_uuid == self._engine.uuid:
            return None
        elif node_uuid in self._remote_nodes:
            return self._remote_nodes[node_uuid]
        else:
            raise ValueError(f"Node with UUID {node_uuid} not found.")

    def resolve_role_conn(
            self,
            role_name: str,
            prefer_self: bool = True
    ) -> tuple[str, aiohttp.ClientWebSocketResponse | web.WebSocketResponse | None]:
        if prefer_self and role_name in self.local_roles:
            return self._engine.uuid, None
        elif role_name in self.remote_roles and len(self.remote_roles[role_name]) > 0:
            node_uuid = list(self.remote_roles[role_name])[0]
            return node_uuid, self.resolve_node_conn(node_uuid)
        else:
            raise ValueError(f"No nodes with Role {role_name} found.")

    def resolve_device_conn(
            self,
            device_uuid: str,
    ) -> tuple[str, aiohttp.ClientWebSocketResponse | web.WebSocketResponse | None]:
        if device_uuid in self.local_devices:
            return self._engine.uuid, None
        elif role_node := self._remote_devices.get(device_uuid):
            return role_node[1], self._remote_nodes[role_node[1]]
        else:
            raise ValueError(f"Device with UUID {device_uuid} not found.")

    def resolve_conn(
            self,
            uuid_or_name: str,
            prefer_self: bool = True,
    ) -> tuple[str, aiohttp.ClientWebSocketResponse | web.WebSocketResponse | None]:
        if uuid_or_name in self.nodes:
            return self.resolve_device_conn(uuid_or_name)
        elif uuid_or_name in self.roles:
            return self.resolve_role_conn(uuid_or_name, prefer_self)
        elif uuid_or_name in self.devices:
            return self.resolve_device_conn(uuid_or_name)
        else:
            raise ValueError(f"No Nodes, Roles, or Devices found with the UUID or name of {uuid_or_name}")

    def get_packet_destination(self, packet: IPCPacket) -> IPCEngine | Role | Device:
        if packet.destination_type is IPCClassType.ENGINE:
            return self._engine
        elif packet.destination_type is IPCClassType.ROLE:
            return self._engine.roles[packet.destination_name]
        elif packet.destination_type is IPCClassType.DEVICE:
            return self._engine.devices[packet.destination_name]
        else:
            raise ValueError(f"Cannot handle destination type %s.", packet.destination_type)

    def get_destination_info_of(self, uuid_or_name: str) -> tuple[
        aiohttp.ClientWebSocketResponse | web.WebSocketResponse | None,
        str,
        IPCClassType,
        str
    ]:
        if uuid_or_name in self.nodes:
            conn = self.resolve_node_conn(uuid_or_name)
            dest_node = uuid_or_name
            dest_type = IPCClassType.ENGINE
            dest_name = uuid_or_name
        elif uuid_or_name in self.roles:
            dest_node, conn = self.resolve_role_conn(uuid_or_name)
            dest_type = IPCClassType.ROLE
            dest_name = uuid_or_name
        elif uuid_or_name in self.devices:
            dest_node, conn = self.resolve_device_conn(uuid_or_name)
            dest_type = IPCClassType.DEVICE
            dest_name = uuid_or_name
        else:
            raise ValueError(f"No Nodes, Roles, or Devices found with the UUID or name of {uuid_or_name}")

        return conn, dest_node, dest_type, dest_name

    def raw_conn_to(
            self,
            requestor: IPCEngine | Role | Device,
            uuid_or_name: str
    ) -> IPCRawConnection:

        conn, dest_node, dest_type, dest_name = self.get_destination_info_of(uuid_or_name)
        origin_type, origin_name, origin_role = get_requestor_info(requestor)

        ret = IPCRawConnection(
            engine=self._engine,
            conn=conn,
            origin_type=origin_type,
            origin_name=origin_name,
            origin_role=origin_role,
            dest_node=dest_node,
            dest_type=dest_type,
            dest_name=dest_name
        )
        return ret

    def routed_conn_to(
            self,
            requestor: IPCEngine | Role | Device,
            uuid_or_name: str,
            dest_chat_uuid: str | None = None
    ):
        conn, dest_node, dest_type, dest_name = self.get_destination_info_of(uuid_or_name)
        ret = IPCRoutedConnection(
            requestor=requestor,
            conn=conn,
            dest_node=dest_node,
            dest_type=dest_type,
            dest_name=dest_name,
            dest_chat_uuid=dest_chat_uuid,
        )
        return ret


class IPCEngine(IPCCore):
    def __init__(self, uuid_override: str | None = None):
        super().__init__()
        self._uuid: str = uuid_override or uuid.uuid1().hex
        self._roles: dict[str, Role] = {}
        self.map: ConnectionMap = ConnectionMap(self)
        self.session: aiohttp.ClientSession | None = None
        self.running: asyncio.Event = asyncio.Event()

        self.events.add_listener(self.on_engine_ready, EngineEvents.ENGINE_READY)
        self.events.add_listener(self.on_engine_closing, EngineEvents.ENGINE_CLOSING)

        self.events.add_listener(self.on_ws_node_added, EngineEvents.WS_NODE_ADDED)
        self.events.add_listener(self.on_node_added, EngineEvents.NODE_ADDED)
        self.events.add_listener(self.on_ws_node_removed, EngineEvents.WS_NODE_REMOVED)
        self.events.add_listener(self.on_node_removed, EngineEvents.NODE_REMOVED)

        self.events.add_listener(self.on_external_role_added, EngineEvents.EXTERNAL_ROLE_ADDED)
        self.events.add_listener(self.on_local_role_added, EngineEvents.LOCAL_ROLE_ADDED)
        self.events.add_listener(self.on_role_added, EngineEvents.ROLE_ADDED)
        self.events.add_listener(self.on_external_role_removed, EngineEvents.EXTERNAL_ROLE_REMOVED)
        self.events.add_listener(self.on_local_role_removed, EngineEvents.LOCAL_ROLE_REMOVED)
        self.events.add_listener(self.on_role_removed, EngineEvents.ROLE_REMOVED)

        self.events.add_listener(self.on_external_device_added, EngineEvents.EXTERNAL_DEVICE_ADDED)
        self.events.add_listener(self.on_local_device_added, EngineEvents.LOCAL_DEVICE_ADDED)
        self.events.add_listener(self.on_device_added, EngineEvents.DEVICE_ADDED)
        self.events.add_listener(self.on_external_device_removed, EngineEvents.EXTERNAL_DEVICE_REMOVED)
        self.events.add_listener(self.on_local_device_removed, EngineEvents.LOCAL_DEVICE_REMOVED)
        self.events.add_listener(self.on_device_removed, EngineEvents.DEVICE_REMOVED)

        self.events.add_listener(self.on_ws_packet, EngineEvents.WS_PACKET)
        self.events.add_listener(self.on_local_packet, EngineEvents.LOCAL_PACKET)
        self.events.add_listener(self.on_packet, EngineEvents.PACKET)
        # self.events.add_listener(self.on_communication, EngineEvents.COMMUNICATION)

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def roles(self) -> dict[str, Role]:
        return self._roles.copy()

    @property
    def devices(self) -> dict[str, Device]:
        ret = {}
        for role in self.roles.values():
            for dev_uuid, device in role.local_devices.items():
                ret[dev_uuid] = device

        return ret

    @property
    def engine(self) -> IPCEngine:
        return self

    def discovery_payload(self) -> IPCPacket:
        ret = IPCPacket(
            payload_type=IPCPayloadType.DISCOVERY,
            origin_type=IPCClassType.ENGINE,
            origin_name=self.uuid,
            origin_role=None,
            dest_type=IPCClassType.ENGINE,
            dest_name=None,
            data=self.uuid,
        )
        return ret

    async def run_stoppable_task(self, coroutine: Coroutine, timeout=0.1) -> asyncio.Task:
        loop = asyncio.get_running_loop()
        task = loop.create_task(coroutine)
        while self.running.is_set():
            try:
                await self.events.wait_for(EngineEvents.ENGINE_CLOSING, timeout=timeout)
            except asyncio.TimeoutError:
                pass
            else:
                logger.debug("Engine is closing, canceling the task.")
                task.cancel()
                break

            if task.done():
                logger.debug("Task %s finished, ending the loop.", coroutine.__name__)
                break

        return task

    async def quick_send_to(self, uuid_or_name: str, packet: IPCPacket):
        node_uuid, conn = self.map.resolve_conn(uuid_or_name)
        if conn is None:
            self.events.dispatch(EngineEvents.LOCAL_PACKET, packet)
        else:
            await conn.send_json(packet.to_dict())

    async def broadcast_to_nodes(self, packet: IPCPacket):
        for node_uuid, ws in self.map.remote_nodes.items():
            packet.destination_type = IPCClassType.ENGINE
            packet.destination_name = node_uuid
            await ws.send_json(packet.to_dict())

    async def update_node(self, node_uuid: str):
        # This prevents roles/devices that are added while running through this from breaking anything.
        current_roles = self.roles
        current_devices = self.devices
        # async with ipc_conn as conn:
        async with self.map.raw_conn_to(self, node_uuid) as conn:
            for role_name in current_roles:
                packet = IPCPacket(
                    payload_type=IPCPayloadType.ROLE_ADD,
                    origin_type=IPCClassType.ENGINE,
                    origin_name=self.uuid,
                    origin_role=None,
                    dest_type=IPCClassType.ENGINE,
                    dest_name=node_uuid,
                    data=role_name,
                )
                await conn.send_packet(packet)

            for device_uuid, device in current_devices.items():
                packet = IPCPacket(
                    payload_type=IPCPayloadType.DEVICE_ADD,
                    origin_type=IPCClassType.ENGINE,
                    origin_name=self.uuid,
                    origin_role=None,
                    dest_type=IPCClassType.ENGINE,
                    dest_name=node_uuid,
                    data={"uuid": device_uuid, "role": device.role.name},
                )
                await conn.send_packet(packet)

    def update_self(self):
        current_roles = self.roles.values()
        current_devices = self.devices.values()

        for role in current_roles:
            self.events.dispatch(EngineEvents.LOCAL_ROLE_ADDED, role)

        for device in current_devices:
            self.events.dispatch(EngineEvents.LOCAL_DEVICE_ADDED, device)

    async def ipc_ws_connect(self, url: str) -> aiohttp.ClientWebSocketResponse | None:
        ws = None
        logger.debug("Starting attempt(s) to establish WS IPC connection to %s", url)
        while not self.session.closed:
            ws = None
            discovered_node = False
            node_uuid = None
            try:
                ws = await self.session.ws_connect(url=url, heartbeat=5)
                await ws.send_json(self.discovery_payload().to_dict())
                async for message in ws:
                    if isinstance(message, aiohttp.WSMessage):
                        if message.type is aiohttp.WSMsgType.text:
                            packet = IPCPacket.from_dict(message.json())
                            if packet.type == IPCPayloadType.DISCOVERY:
                                node_uuid = packet.data
                                self.map.clean()
                                if discovered_node and node_uuid in self.map.nodes:
                                    logger.warning(
                                        "Outgoing Node %s sent a discovery packet despite already being connected. "
                                        "Ignoring it, but is everything okay?", node_uuid
                                    )
                                if self.map.add_node(ws, node_uuid):
                                    logger.debug("Node %s sent a discovery, dispatching event.", node_uuid)
                                    discovered_node = True
                                    self.events.dispatch(EngineEvents.WS_NODE_ADDED, ws, node_uuid)
                                else:
                                    logger.debug(
                                        "Node %s is already connected to us. Closing outgoing connection.", node_uuid
                                    )
                                    await ws.close()
                                    return ws
                            elif discovered_node is False:
                                logger.debug("Non-discovery JSON message received before a discovery one, discarding.")
                            else:
                                self.events.dispatch(EngineEvents.WS_PACKET, ws, packet, node_uuid)
                        else:
                            logger.debug("Unhandled non-text message received, discarding.")

                self.map.clean()

            except (aiohttp.ClientConnectorError, ConnectionRefusedError) as e:
                logger.warning(
                    "Cannot connect or connection lost to %s, waiting %s seconds before retrying.",
                    url,
                    CONNECT_RETRY_SLEEP
                )
                # TODO: More gracefully handle it so _handle_outgoing_ws is also reset?
                await asyncio.sleep(CONNECT_RETRY_SLEEP)

            except aiohttp.WSServerHandshakeError as e:
                match e.status:
                    case 404:
                        logger.error(
                            "Received a 404: %s when connecting to %s, is that route set up for Websockets?",
                            e.message, url
                        )
                        break
                    case _:
                        logger.error(
                            "Received a %s:%s when connecting to %s, no handler available?",
                            e.status, e.message, url
                        )

            except RuntimeError as e:
                logger.exception(
                    "Encountered runtime error while connected to %s, waiting %s seconds before retrying.",
                    url,
                    CONNECT_RETRY_SLEEP,
                    exc_info=e
                )
                await asyncio.sleep(CONNECT_RETRY_SLEEP)

        return ws

    def ipc_middleware(self, route: str = f"/{ENGINE_IPC_ROUTE}"):
        @web.middleware
        async def engine_ipc_middleware(request: web.Request, handler):
            if request.path == route and request.method == "GET":
                return await self.run_stoppable_task(self._ws_ipc_receive_handler(request), timeout=20)
            else:
                return await handler(request)

        return engine_ipc_middleware

    async def _ws_ipc_receive_handler(self, request: web.Request) -> web.WebSocketResponse:
        logger.debug("Incoming WS IPC connection from %s", request.remote)
        discovered_node = False
        node_uuid = None
        ws = web.WebSocketResponse(heartbeat=5)
        await ws.prepare(request)

        await ws.send_json(self.discovery_payload().to_dict())
        async for message in ws:
            if isinstance(message, aiohttp.WSMessage):
                if message.type is aiohttp.WSMsgType.text:
                    packet = IPCPacket.from_dict(message.json())
                    if packet.type == IPCPayloadType.DISCOVERY:
                        node_uuid = packet.data
                        self.map.clean()
                        if self.map.add_node(ws, node_uuid):
                            if discovered_node and node_uuid in self.map.nodes:
                                logger.warning(
                                    "Incoming Node %s sent a discovery packet despite already being connected. "
                                    "Ignoring it, but is everything okay?", node_uuid
                                )
                            else:
                                logger.debug("Incoming Node %s sent a discovery, dispatching event.", node_uuid)
                                discovered_node = True
                                self.events.dispatch(EngineEvents.WS_NODE_ADDED, ws, node_uuid)
                        else:
                            logger.debug(
                                "Incoming Node %s is already connected to us, closing incoming connection.", node_uuid
                            )
                            await ws.close()
                            return ws
                    elif not discovered_node:
                        logger.debug("Non-discovery JSON message was sent before a discovery one, ignoring.")
                    else:
                        self.events.dispatch(EngineEvents.WS_PACKET, ws, packet, node_uuid)

        self.map.clean()

        return ws

    async def on_engine_ready(self):
        logger.debug("IPC Engine is ready.")

    async def on_engine_closing(self):
        logger.debug("IPC Engine is closing.")
        self.running.clear()

    async def on_ws_node_added(self, ws: web.WebSocketResponse | aiohttp.ClientWebSocketResponse, node_uuid: str):
        self.map.add_node(ws, node_uuid)
        self.events.dispatch(EngineEvents.NODE_ADDED, node_uuid)

    async def on_node_added(self, node_uuid: str):
        logger.info("Node %s added.", node_uuid)
        await self.update_node(node_uuid)

    async def on_ws_node_removed(self, ws: web.WebSocketResponse | aiohttp.ClientWebSocketResponse, node_uuid: str):
        self.events.dispatch(EngineEvents.NODE_REMOVED, node_uuid)

    async def on_node_removed(self, node_uuid: str):
        logger.info("Node removed: %s", node_uuid)

    async def on_external_role_added(
            self,
            packet: IPCPacket,
            node_uuid: str
    ):
        self.map.add_external_role(packet.data, node_uuid)
        self.events.dispatch(EngineEvents.ROLE_ADDED, packet.data, node_uuid)

    async def on_local_role_added(self, role: Role):
        self.events.dispatch(EngineEvents.ROLE_ADDED, role.name, self.uuid)
        if self.session:
            packet = IPCPacket(
                payload_type=IPCPayloadType.ROLE_ADD,
                origin_type=IPCClassType.ENGINE,
                origin_name=self.uuid,
                origin_role=None,
                dest_type=IPCClassType.ENGINE,
                dest_name=None,
                data=role.name
            )
            await self.broadcast_to_nodes(packet)

    async def on_role_added(self, role_name: str, node_uuid: str):
        logger.debug("New Role %s from Node %s", role_name, node_uuid)

    async def on_external_role_removed(
            self,
            packet: IPCPacket | str,
            node_uuid: str
    ):
        if isinstance(packet, IPCPacket):
            role_name = packet.data
        elif isinstance(packet, str):
            role_name = packet
        else:
            logger.warning("Unhandled type %s for packet variable", type(packet))
            return

        self.events.dispatch(EngineEvents.ROLE_REMOVED, role_name, node_uuid)

    async def on_local_role_removed(self, role: Role):
        self.events.dispatch(EngineEvents.ROLE_REMOVED, role.name, self.uuid)

    async def on_role_removed(self, role_name: str, node_uuid: str):
        logger.debug("Node %s removed role %s.", node_uuid, role_name)

    async def on_external_device_added(self, packet: IPCPacket, node_uuid: str):
        device_uuid = packet.data["uuid"]
        device_role = packet.data["role"]
        logger.debug("Node %s informed us of device %s with role %s, adding.", node_uuid, device_uuid, device_role)
        self.map.add_external_device(device_uuid, device_role, node_uuid)
        self.events.dispatch(EngineEvents.DEVICE_ADDED, device_uuid, device_role, node_uuid)

    async def on_local_device_added(self, device: Device):
        self.events.dispatch(EngineEvents.DEVICE_ADDED, device.uuid, device.role.name, self.uuid)
        if self.session:
            packet = IPCPacket(
                payload_type=IPCPayloadType.DEVICE_ADD,
                origin_type=IPCClassType.ENGINE,
                origin_name=self.uuid,
                origin_role=None,
                dest_type=IPCClassType.ENGINE,
                dest_name=None,
                data={"uuid": device.uuid, "role": device.role.name}
            )
            await self.broadcast_to_nodes(packet)

    async def on_device_added(self, device_uuid: str, device_role: str, node_uuid: str):
        logger.debug("New device %s with role %s from node %s added.", device_uuid, device_role, node_uuid)

    async def on_external_device_removed(self, packet: IPCPacket | str, node_uuid: str | tuple[str, str]):
        if isinstance(packet, IPCPacket) and isinstance(node_uuid, str):
            device_role = packet.data
            self.events.dispatch(EngineEvents.DEVICE_REMOVED, device_role[0], device_role[1], node_uuid)
        elif isinstance(packet, str) and isinstance(node_uuid, tuple):
            self.events.dispatch(EngineEvents.DEVICE_REMOVED, packet, node_uuid[0], node_uuid[1])
        else:
            logger.warning("Unhandled packet and node_uuid type combination: %s %s", type(packet), type(node_uuid))

    async def on_local_device_removed(self, device: Device):
        self.events.dispatch(EngineEvents.DEVICE_REMOVED, device.uuid, device.role.name, self.uuid)

    async def on_device_removed(self, device_uuid: str, device_role: str, node_uuid: str):
        logger.debug("Node %s removed device %s with role %s.", node_uuid, device_uuid, device_role)

    async def on_ws_packet(
            self,
            ws: web.WebSocketResponse | aiohttp.ClientWebSocketResponse,
            packet: IPCPacket,
            node_uuid: str
    ):
        self.events.dispatch(EngineEvents.PACKET, packet, node_uuid)

    async def on_local_packet(self, packet: IPCPacket):
        self.events.dispatch(EngineEvents.PACKET, packet, self.uuid)

    async def on_packet(self, packet: IPCPacket, node_uuid: str):
        match packet.type:
            case IPCPayloadType.COMMUNICATION:
                dest = self.map.get_packet_destination(packet)
                logger.debug("Chat communication received, routing to %s", dest)
                if packet.event is None or packet.dest_conn_uuid is not None:
                    dest.events.dispatch(CoreEvents.ROUTED_CONN_MESSAGE, packet, node_uuid)
                else:
                    dest.events.dispatch(packet.event, packet, node_uuid)
            case IPCPayloadType.COMMUNICATION_REQUEST:
                dest = self.map.get_packet_destination(packet)
                logger.debug("Chat communication request received, routing to %s", dest)
                dest.events.dispatch(CoreEvents.ROUTED_CONN_INCOMING, packet, node_uuid)
                # TODO: Put in logic to route to the correct place.
            case IPCPayloadType.COMMUNICATION_ACCEPTED:
                dest = self.map.get_packet_destination(packet)
                logger.debug("Chat communication acceptance received, routing to %s", dest)
                dest.events.dispatch(CoreEvents.ROUTED_CONN_MESSAGE, packet, node_uuid)
                # TODO: Put in logic to route to the correct place.
            case IPCPayloadType.COMMUNICATION_DENIED:
                dest = self.map.get_packet_destination(packet)
                logger.debug("Chat communication denial received, routing to %s", dest)
                dest.events.dispatch(CoreEvents.ROUTED_CONN_MESSAGE, packet, node_uuid)
                # TODO: Put in logic to route to the correct place.
            case IPCPayloadType.COMMUNICATION_REDIRECT:
                dest = self.map.get_packet_destination(packet)
                logger.debug("Chat communication redirect received, routing to %s", dest)
                dest.events.dispatch(CoreEvents.ROUTED_CONN_MESSAGE, packet, node_uuid)
                # TODO: Put in logic to route to the correct place.
            case IPCPayloadType.DISCOVERY:
                logger.warning("We aren't supposed to be able to handle discovery here?")
            case IPCPayloadType.ROLE_ADD:
                self.events.dispatch(EngineEvents.EXTERNAL_ROLE_ADDED, packet, node_uuid)
            case IPCPayloadType.ROLE_REMOVE:
                self.events.dispatch(EngineEvents.EXTERNAL_ROLE_REMOVED, packet, node_uuid)
            case IPCPayloadType.DEVICE_ADD:
                self.events.dispatch(EngineEvents.EXTERNAL_DEVICE_ADDED, packet, node_uuid)
            case IPCPayloadType.DEVICE_REMOVE:
                self.events.dispatch(EngineEvents.EXTERNAL_DEVICE_REMOVED, packet, node_uuid)
            case _:
                logger.warning(
                    "Unknown WS IPC message type encountered from node %s: %s", node_uuid, packet.type
                )

    def add_role(self, role: Role):
        if role.name in self._roles:
            raise ValueError(f"A role with name {role.name} has already been added.")

        self._roles[role.name] = role
        role.set_engine(self)
        if self.running.is_set():
            packet = IPCPacket(
                payload_type=IPCPayloadType.ROLE_ADD,
                origin_type=IPCClassType.ENGINE,
                origin_name=self.uuid,
                origin_role=None,
                dest_type=IPCClassType.ENGINE,
                dest_name=None,
                data=role.name,
            )
            loop = asyncio.get_running_loop()
            loop.create_task(self.broadcast_to_nodes(packet))

    async def start(self):
        self.running.set()
        self.events.dispatch(EngineEvents.ENGINE_READY)
        self.update_self()
        logger.debug("IPC Engine started.")

    async def close(self, closing_time: float = 1.0):
        self.events.dispatch(EngineEvents.ENGINE_CLOSING)
        await self.session.close()

        await asyncio.sleep(closing_time)

    async def start_serverless(self, *, discover_nodes: list[tuple[str, int]] | None) -> None:
        self.session = aiohttp.ClientSession()
        logger.info("%s starting serverless.", self.__class__.__name__)
        await self.start()
        if discover_nodes:
            loop = asyncio.get_running_loop()
            for node_address, node_port in discover_nodes:
                loop.create_task(self.run_stoppable_task(
                    self.ipc_ws_connect(f"http://{node_address}:{node_port}/{ENGINE_IPC_ROUTE}")
                ))

    def run_serverless(
            self,
            *,
            loop: asyncio.AbstractEventLoop | None = None,
            closing_time: float = 1.0,
            discover_nodes: list[tuple[str, int]] | None
    ):
        loop = loop or asyncio.new_event_loop()
        task = loop.create_task(self.start_serverless(discover_nodes=discover_nodes))
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            logger.debug("KeyboardInterrupt encountered, stopping loop.")
            logger.debug("Cancelling start task.")
            task.cancel()

            loop.run_until_complete(self.close(closing_time))

    async def start_server(
            self,
            *,
            port: int = 8080,
            discover_nodes: list[tuple[str, int]] | None = None,
    ) -> web.BaseSite:
        self.session = aiohttp.ClientSession()
        app = web.Application(middlewares=[self.ipc_middleware()])
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, "0.0.0.0", port=port, shutdown_timeout=5.0)
        await site.start()
        logger.info("%s listening on %s.", self.__class__.__name__, site.name)
        await self.start()
        if discover_nodes:
            loop = asyncio.get_running_loop()
            for node_address, node_port in discover_nodes:
                loop.create_task(self.run_stoppable_task(
                    self.ipc_ws_connect(f"http://{node_address}:{node_port}/{ENGINE_IPC_ROUTE}")
                ))

        return site

    def run_server(
            self,
            *,
            loop: asyncio.AbstractEventLoop | None = None,
            port: int = 8080,
            closing_time: float = 1.0,
            discover_nodes: list[tuple[str, int]] | None = None,
    ):
        def on_interrupt():
            logger.info("Stop signal encountered, stopping event loop.")
            loop.stop()

        loop = loop or asyncio.new_event_loop()

        for sig in ON_SIGNAL_CLOSE:
            loop.add_signal_handler(sig, on_interrupt)

        task = loop.create_task(self.start_server(port=port, discover_nodes=discover_nodes))
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            logger.debug("KeyboardInterrupt encountered, stopping loop.")
        except Exception as e:
            logger.error("Exception occurred that broke out of the event loop, wtf?", exc_info=e)

        if site := task.result():
            logger.debug("Site was created, attempting to stop it.")
            loop.run_until_complete(site.stop())
        else:
            logger.debug("Cancelling start task.")
            task.cancel()

        loop.run_until_complete(self.close(closing_time))



from __future__ import annotations

from aiohttp import web
from logging import getLogger
from typing import AsyncIterator, TYPE_CHECKING

import aiohttp
import asyncio
import uuid

from .enums import CoreEvents, EngineEvents, IPCClassType, IPCPayloadType
from .core import DispatchFramework


if TYPE_CHECKING:
    from .engine import IPCEngine
    from .role import Role
    from .device import Device


__all__ = (
    "get_requestor_info",
    "IPCRawConnection",
    "IPCRoutedConnection",
    "IPCCore",
    "IPCPacket",
)


logger = getLogger(__name__)


ConnDataTypes = str | dict | bytes | None


class IPCPacket:
    def __init__(
            self,
            *,
            payload_type: IPCPayloadType,
            origin_type: IPCClassType,
            origin_name: str | None,
            origin_role: str | None,
            dest_type: IPCClassType,
            dest_name: str | None,
            data: ConnDataTypes,
            origin_conn_uuid: str | None = None,
            dest_conn_uuid: str | None = None,
            event: str | None = None,
            packet_uuid: str | None = None,
            is_response: bool = False,
    ):
        self.type = payload_type
        self.origin_type = origin_type
        """IPC Class Type that the packet is being sent from."""
        self.origin_name = origin_name
        """Role name for Role origin, Device UUID for Device origin, node UUID for Engine origin."""
        self.origin_role = origin_role
        """Role name of the Device for Device origin. None for Role and Engine origin."""
        self.destination_type = dest_type
        """IPC Class Type that the packet is being sent to."""
        self.destination_name = dest_name
        """Role name for Role origin, Device UUID for Device origin, node UUID for Engine origin."""
        self.data = data
        """Actual data to be transmitted in the packet."""
        self.origin_conn_uuid = origin_conn_uuid
        """Used to indicate what connection a specific packet is from."""
        self.dest_conn_uuid = dest_conn_uuid
        """UUID of the target connection, typically used to indicate that the specific connection should be modified."""
        self.event = event
        """Event that the destination should broadcast with the packet as the arg."""

        self._uuid = packet_uuid
        """UUID of the packet. Should only be sent if this packet is requesting or responding."""
        self._is_response = is_response
        """Indicates if this packet is a new request or a response to an existing request."""

        self._conn = None

    @property
    def is_request(self) -> bool:
        return self._uuid is not None and self.is_response is False

    @property
    def uuid(self) -> str | None:
        return self._uuid

    @property
    def is_response(self) -> bool:
        return self._is_response

    def set_connection(self, conn: IPCRawConnection):
        self._conn = conn

    @classmethod
    def from_connection(
            cls,
            conn: IPCRawConnection,
            payload_type: IPCPayloadType,
            data: ConnDataTypes,
            *,
            event: str | None = None,
            packet_uuid: str | None = None,
            is_response: bool = False,
    ) -> IPCPacket:
        ret = cls(
            payload_type=payload_type,
            origin_type=conn.origin_type,
            origin_name=conn.origin_name,
            origin_role=conn.origin_role,
            dest_type=conn.dest_type,
            dest_name=conn.dest_name,
            data=data,
            origin_conn_uuid=conn.uuid,
            event=event,
            packet_uuid=packet_uuid,
            is_response=is_response,
        )
        return ret

    def to_dict(self) -> dict:
        ret = {
            "type": self.type.value,
            "origin_type": self.origin_type.value,
            "origin_name": self.origin_name,
            "origin_role": self.origin_role,
            "destination_type": self.destination_type.value,
            "destination_name": self.destination_name,
            "data": self.data,
            "origin_conn_uuid": self.origin_conn_uuid,
            "dest_conn_uuid": self.dest_conn_uuid,
            "event": self.event,
            "_uuid": self._uuid,
            "_is_response": self._is_response,
        }
        return ret

    @classmethod
    def from_dict(cls, packet: dict) -> IPCPacket:
        ret = cls(
            payload_type=IPCPayloadType(packet["type"]),
            origin_type=IPCClassType(packet["origin_type"]),
            origin_name=packet["origin_name"],
            origin_role=packet["origin_role"],
            dest_type=IPCClassType(packet["destination_type"]),
            dest_name=packet["destination_name"],
            data=packet["data"],
            origin_conn_uuid=packet["origin_conn_uuid"],
            dest_conn_uuid=packet["dest_conn_uuid"],
            event=packet["event"],
            packet_uuid=packet["_uuid"],
            is_response=packet["_is_response"]
        )
        return ret

    async def send_response(self, data: ConnDataTypes):
        if self.uuid is None:
            raise ValueError("A response requires the request packet to have a UUID.")
        elif not isinstance(self._conn, IPCRoutedConnection):
            raise ValueError("Responses can only be performed by an IPCChat object.")
        else:
            await self._conn.send_response(data, self.uuid)


class IPCRawConnection:
    class ConnectionClosed(Exception):
        pass

    def __init__(
            self,
            engine: IPCEngine,
            conn: web.WebSocketResponse | aiohttp.ClientWebSocketResponse | None,
            origin_type: IPCClassType,
            origin_name: str,
            origin_role: str | None,
            dest_node: str,
            dest_type: IPCClassType,
            dest_name: str,
            uuid_override: str | None = None,
    ):
        """A "raw" connection to the given destination. Can only send packets, has no receiving capabilities.

        Parameters
        ----------
        engine
        conn
        origin_type
        origin_name
        origin_role
        dest_node
        dest_type
        dest_name
        uuid_override
        """
        self._engine = engine
        self._conn = conn
        self._origin_type = origin_type
        self._origin_name = origin_name
        self._origin_role = origin_role
        self._dest_node = dest_node
        self._dest_type = dest_type
        self._dest_name = dest_name
        self._uuid = uuid_override or uuid.uuid4().hex
        self._is_open: bool = False

    @property
    def origin_type(self) -> IPCClassType:
        return self._origin_type

    @property
    def origin_name(self) -> str:
        return self._origin_name

    @property
    def origin_role(self) -> str | None:
        return self._origin_role

    @property
    def dest_node(self) -> str:
        return self._dest_node

    @property
    def dest_type(self) -> IPCClassType:
        return self._dest_type

    @property
    def dest_name(self) -> str:
        return self._dest_name

    @property
    def dest_role(self) -> str | None:
        if self.dest_type == IPCClassType.DEVICE:
            return self._engine.map.full_device_info[self.dest_name][0]
        else:
            return None

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def is_open(self) -> bool:
        return self._is_open

    @classmethod
    def from_packet(cls, requestor: IPCCore, packet: IPCPacket, node_uuid: str) -> IPCRawConnection:
        """Makes a connection to the origin of the given packet."""
        origin_type, origin_name, origin_role = get_requestor_info(requestor)
        conn = requestor.engine.map.resolve_node_conn(node_uuid)
        ret = cls(
            engine=requestor.engine,
            conn=conn,
            origin_type=origin_type,
            origin_name=origin_name,
            origin_role=origin_role,
            dest_node=node_uuid,
            dest_type=packet.origin_type,
            dest_name=packet.origin_name,
        )
        return ret

    async def open(self) -> bool:
        """Opens the connection, allowing the sending of packets and hooks listeners into the engine."""
        if not self.is_open:
            self._is_open = True
            self._engine.events.add_listener(self._on_engine_close, EngineEvents.ENGINE_CLOSING)
            self._engine.events.add_listener(self._on_node_removed, EngineEvents.NODE_REMOVED)
            return True
        else:
            return False

    async def close(self) -> bool:
        """Closes the connection, preventing sending of packets and unhooking listeners."""
        self._is_open = False
        self._engine.events.remove_listener(self._on_engine_close, EngineEvents.ENGINE_CLOSING)
        self._engine.events.remove_listener(self._on_node_removed, EngineEvents.NODE_REMOVED)
        return True

    async def change_dest(self, dest_node: str, dest_type: IPCClassType, dest_name: str | None):
        self._conn = self._engine.map.resolve_node_conn(dest_node)
        self._dest_node = dest_node
        self._dest_type = dest_type
        self._dest_name = dest_name

    async def _on_engine_close(self):
        await self.close()

    async def _on_node_removed(self, node_uuid: str):
        if node_uuid == self.dest_node:
            await self.close()

    async def send_packet(self, packet: IPCPacket, ignore_checks=False):
        # if not ignore_checks and not self.is_open:
        if ignore_checks is False and self.is_open is False:
            raise self.ConnectionClosed("The IPC connection has been closed.")

        if self._conn is None:
            self._engine.events.dispatch(EngineEvents.LOCAL_PACKET, packet)
        else:
            try:
                await self._conn.send_json(packet.to_dict())
            except Exception as e:
                logger.debug(
                    "%s error when communicating with node %s, closing connection.",
                    e.__class__,
                    self._dest_node
                )
                await self.close()
                raise self.ConnectionClosed("The IPC connection has been closed.")

    async def __aenter__(self):
        if self.is_open is False:
            await self.open()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.is_open:
            await self.close()


class IPCRoutedConnection(IPCRawConnection):
    class CommunicationDenied(Exception):
        pass

    _MESSAGE_REPLY = "MESSAGE_REPLY"

    def __init__(
            self,
            *,
            requestor: IPCCore,
            conn: web.WebSocketResponse | aiohttp.ClientWebSocketResponse | None,
            dest_node: str,
            dest_type: IPCClassType,
            dest_name: str,
            dest_chat_uuid: str | None = None,
            uuid_override: str | None = None
    ):
        """A "routed" connection to the destination, designed for 2-way communication.
        Routed Connections must be accepted by the destination before packets can be sent, and can be denied
        or redirected.

        Parameters
        ----------
        requestor:
            Object requesting this connection to be created.
        conn:
            The specific connection to use for communication to the destination.
        dest_node
        dest_type
        dest_name
        dest_chat_uuid
        uuid_override
        """
        origin_type, origin_name, origin_role = get_requestor_info(requestor)
        super().__init__(
            engine=requestor.engine,
            conn=conn,
            origin_type=origin_type,
            origin_name=origin_name,
            origin_role=origin_role,
            dest_node=dest_node,
            dest_type=dest_type,
            dest_name=dest_name,
            uuid_override=uuid_override
        )
        self._requestor = requestor
        self._dest_chat_uuid: str | None = dest_chat_uuid
        self._ready_for_communication: asyncio.Event = asyncio.Event()
        """If the IPCConnection is ready for the user to attempt to use send_chat. This being set does NOT mean that
        it will be successful, only that the user shouldn't wait for the connection to be set.
        """
        self._sent_chat_request: bool = False
        """If we already sent a chat request. If True, no further communication requests should be sent."""
        self._chat_denied: bool = False
        """If the chat request was denied. If True, no further communication data packets should be sent."""
        self._packet_queue: asyncio.Queue[IPCPacket | None] = asyncio.Queue()
        self._events = DispatchFramework()

    @property
    def dest_chat_uuid(self) -> str:
        return self._dest_chat_uuid

    @classmethod
    def from_packet(cls, requestor: IPCCore, packet: IPCPacket, node_uuid: str) -> IPCRoutedConnection:
        conn = requestor.engine.map.resolve_node_conn(node_uuid)
        ret = cls(
            requestor=requestor,
            conn=conn,
            dest_node=node_uuid,
            dest_type=packet.origin_type,
            dest_name=packet.origin_name,
            dest_chat_uuid=packet.origin_conn_uuid
        )

        return ret

    async def wait_until_comm_ready(self):
        """Blocking call that waits until communication is ready."""
        await self._ready_for_communication.wait()
        if self._chat_denied:
            raise self.CommunicationDenied("Communication request has been denied.")

    async def open(self) -> bool:
        """Opens the connection, """
        if self._chat_denied:
            raise self.CommunicationDenied("Communication request has been denied.")

        if await super().open():
            self._is_open = False
            self._requestor.events.add_listener(self.on_ipc_message, CoreEvents.ROUTED_CONN_MESSAGE)
            if self._sent_chat_request is False:
                await self.send_chat_request()

            return True
        else:
            return False

    async def close(self, force: bool = False) -> bool:
        self._is_open = False
        self._requestor.events.remove_listener(self.on_ipc_message, CoreEvents.ROUTED_CONN_MESSAGE)
        self._ready_for_communication.clear()
        await self._packet_queue.put(None)
        return True

    async def send_chat_request(self):
        """Sends an outgoing chat request, asking permission to communicate."""
        comm_request = IPCPacket.from_connection(self, IPCPayloadType.COMMUNICATION_REQUEST, None)
        comm_request.dest_conn_uuid = self.dest_chat_uuid
        await self.send_packet(comm_request, ignore_checks=True)
        self._sent_chat_request = True

    async def accept_chat_request(self, dest_chat_uuid: str):
        """Accepts the incoming chat request, signalling that we are ready to communicate."""
        packet = IPCPacket.from_connection(self, IPCPayloadType.COMMUNICATION_ACCEPTED, dest_chat_uuid)
        packet.dest_conn_uuid = dest_chat_uuid
        self._dest_chat_uuid = dest_chat_uuid
        self._sent_chat_request = True
        await self.open()
        self._is_open = True
        await self.send_packet(packet)
        self._ready_for_communication.set()

    async def deny_chat_request(self, dest_chat_uuid: str):
        """Denies the incoming chat request, then closes the connection on this side."""
        packet = IPCPacket.from_connection(self, IPCPayloadType.COMMUNICATION_DENIED, dest_chat_uuid)
        packet.dest_conn_uuid = dest_chat_uuid
        self._dest_chat_uuid = dest_chat_uuid
        self._sent_chat_request = True
        await self.open()
        self._is_open = True
        await self.send_packet(packet)
        self._chat_denied = True
        self._ready_for_communication.set()
        await self.close()

    async def redirect_chat_request(self, dest_type: IPCClassType, dest_name: str, dest_node: str, dest_chat_uuid: str):
        """Redirects the incoming chat request to a new destination, then closes the connection on this side."""
        payload = {
            "destination_type": dest_type.value, "destination_name": dest_name, "destination_node": dest_node
        }
        packet = IPCPacket.from_connection(self, IPCPayloadType.COMMUNICATION_REDIRECT, payload)
        packet.dest_conn_uuid = dest_chat_uuid
        self._dest_chat_uuid = dest_chat_uuid
        self._sent_chat_request = True
        await self.open()
        self._is_open = True
        await self.send_packet(packet)
        self._chat_denied = True
        self._ready_for_communication.set()
        await self.close()

    async def send_data(
            self,
            data: ConnDataTypes,
            event: str | None = None,
            *,
            set_dest: bool = True,
            payload_type: IPCPayloadType = IPCPayloadType.COMMUNICATION
    ):
        """Makes and sends a packet with the given data."""
        packet = IPCPacket.from_connection(self, payload_type, data=data, event=event)
        if set_dest:
            packet.dest_conn_uuid = self.dest_chat_uuid

        await self.send_packet(packet)

    async def send_request(
            self,
            data: ConnDataTypes,
            event: str | int | None = None,
            timeout: float = 5.0
    ) -> IPCPacket:
        """Sends a communication data packet, expecting a response from it."""
        packet_uuid = uuid.uuid1().hex
        packet = IPCPacket.from_connection(
            self,
            IPCPayloadType.COMMUNICATION,
            data,
            event=event,
            packet_uuid=packet_uuid
        )
        packet.dest_conn_uuid = self.dest_chat_uuid
        await self.send_packet(packet)
        # logger.critical("WAITING FOR P.UUID TO EQUAL %s", packet_uuid)
        return await self._events.wait_for(self._MESSAGE_REPLY, lambda p: p.uuid == packet_uuid, timeout=timeout)

    async def send_response(self, data: ConnDataTypes, request_packet_uuid: str):
        """Sends a response packet with the given data, responding to the given packet UUID."""
        packet = IPCPacket.from_connection(
            self, IPCPayloadType.COMMUNICATION, data, packet_uuid=request_packet_uuid, is_response=True
        )
        packet.dest_conn_uuid = self.dest_chat_uuid
        await self.send_packet(packet)

    async def receive(self) -> IPCPacket | None:
        """Waits until either a new packet arrives or the IPC connection closes."""
        if self._conn is not None and self._conn.closed:
            raise self.ConnectionClosed("The IPC connection has been closed.")

        ret = await self._packet_queue.get()
        self._packet_queue.task_done()
        if ret is None:
            raise self.ConnectionClosed("The IPC connection has been closed.")

        return ret

    async def __aenter__(self):
        await super().__aenter__()
        await self.wait_until_comm_ready()
        return self

    def __aiter__(self) -> AsyncIterator[IPCPacket]:
        return self

    async def __anext__(self) -> IPCPacket:
        if self.is_open:
            try:
                return await self.receive()
            except self.ConnectionClosed:
                raise StopAsyncIteration

        else:
            raise StopAsyncIteration

    async def on_ipc_message(self, packet: IPCPacket, origin_node: str | None):
        if packet.dest_conn_uuid == self.uuid:
            packet.set_connection(self)
            if self.is_open:
                if packet.type is IPCPayloadType.COMMUNICATION:
                    if packet.is_response:
                        self._events.dispatch(self._MESSAGE_REPLY, packet)
                    elif packet.event:
                        self._requestor.events.dispatch(packet.event, packet)
                    else:
                        await self._packet_queue.put(packet)
            elif self._chat_denied is False:
                match packet.type:
                    case IPCPayloadType.COMMUNICATION_ACCEPTED:
                        self._is_open = True
                        self._dest_chat_uuid = packet.origin_conn_uuid
                        self._ready_for_communication.set()
                    case IPCPayloadType.COMMUNICATION_DENIED:
                        self._chat_denied = True
                        self._ready_for_communication.set()
                        await self.close(force=True)
                    case IPCPayloadType.COMMUNICATION_REDIRECT:
                        dest_type = IPCClassType(packet.data["destination_type"])
                        dest_name = packet.data["destination_name"]
                        dest_node = packet.data["destination_node"]
                        await self.change_dest(dest_node, dest_type, dest_name)
                        await self.send_chat_request()


class IPCCore:
    engine: IPCEngine

    def __init__(self):
        """Contains the core bits required for IPC shared between the Engine, Role, and Device."""
        self.events = DispatchFramework()
        self.events.add_listener(self.on_incoming_chat, CoreEvents.ROUTED_CONN_INCOMING)
        self.events.add_listener(self.on_chat_connection, CoreEvents.ROUTED_CONN_CONNECTION)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"

    async def on_incoming_chat(self, packet: IPCPacket, node_uuid: str):
        """Ran when an incoming connection is established but needs to be accepted."""
        logger.debug("Received incoming connection from node %s.", node_uuid)
        if await self.accept_incoming_chat(packet, node_uuid):
            logger.debug("Incoming connection locally accepted. Creating conn, sending accept, and dispatching.")
            conn = IPCRoutedConnection.from_packet(self, packet, node_uuid)
            await conn.accept_chat_request(packet.origin_conn_uuid)
            self.events.dispatch(CoreEvents.ROUTED_CONN_CONNECTION, conn)
        else:
            logger.debug("Incoming connection locally denied. Creating conn, sending deny, and ignoring.")
            conn = IPCRoutedConnection.from_packet(self, packet, node_uuid)
            await conn.deny_chat_request(packet.origin_conn_uuid)

    # noinspection PyMethodMayBeStatic
    async def accept_incoming_chat(self, packet: IPCPacket, node_uuid: str) -> bool:
        """Used for custom logic for if a connection should be accepted or not. Return True to accept it, False to
        deny the connection.
        """
        return True

    async def on_chat_connection(self, chat: IPCRoutedConnection):
        """Ran when an incoming connection is established and accepted."""
        await chat.close()
        raise NotImplementedError


# Importing all the way down here helps prevent import errors.
from . import engine, role, device


def get_requestor_info(requestor: IPCEngine | Role | Device) -> tuple[IPCClassType, str, str | None]:
    """Grabs the origin type, name, and role from the given requestor."""
    if isinstance(requestor, engine.IPCEngine):
        origin_type = IPCClassType.ENGINE
        origin_name = requestor.uuid
        origin_role = None
    elif isinstance(requestor, role.Role):
        origin_type = IPCClassType.ROLE
        origin_name = requestor.name
        origin_role = None
    elif isinstance(requestor, device.Device):
        origin_type = IPCClassType.DEVICE
        origin_name = requestor.uuid
        origin_role = requestor.role.name
    else:
        raise ValueError("Requester type %s is not supported.", type(requestor))

    return origin_type, origin_name, origin_role



from setuptools import setup, find_packages
import pathlib


with open("requirements.txt") as f:
    requirements = f.read().splitlines()


here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")


setup(
    name="mycometo",
    version="0.0.4",
    description="A multi-layer IPC framework for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alentoghostflame/mycometo",
    author="Alex Schoenhofen",
    author_email="alexanderschoenhofen@gmail.com",
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="ipc, development",
    packages=find_packages(exclude="examples"),
    install_requires=requirements,
)



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING


from .enums import EngineEvents, IPCClassType
from .connection import IPCRoutedConnection
from . import connection, device


if TYPE_CHECKING:
    from .engine import IPCEngine
    from .device import Device


__all__ = (
    "Role",
    "SingleDeviceRole",
)


logger = getLogger(__name__)


class Role(connection.IPCCore):
    def __init__(self, name: str):
        super().__init__()
        self._name: str = name
        self._engine: IPCEngine | None = None
        """Set with the Role is added to an Engine."""
        self._devices: dict[str, Device] = {}
        """{"Device UUID": Device}"""

    @property
    def name(self) -> str:
        return self._name

    @property
    def engine(self) -> IPCEngine | None:
        return self._engine

    @property
    def local_devices(self) -> dict[str, Device]:
        return self._devices.copy()

    @property
    def devices(self) -> list[str]:
        """Returns a list of device UUIDs that have this as their role. Includes remote devices."""
        return [device for device, role_node in self.engine.map.devices if role_node[0] == self.name]

    def set_engine(self, engine: IPCEngine):
        self._engine = engine
        for device in self._devices.values():
            if device.engine is None:
                device.set_engine(engine)

        self.on_set_engine(engine)

    def on_set_engine(self, engine: IPCEngine):
        """Ran when the engine is set. Exists for the sole purpose of being optionally overridden."""
        pass

    def add_device(self, device: Device):
        if device.uuid in self._devices and self._devices[device.uuid] is not device:
            raise ValueError(f"Device with UUID {device.uuid} has already been added to this role.")

        self._devices[device.uuid] = device
        if self._engine is not None:
            device.set_engine(self.engine)

        device.set_role(self)
        if self.engine and self.engine.running.is_set():
            self.engine.events.dispatch(EngineEvents.LOCAL_DEVICE_ADDED, device)


class SingleDeviceRole(Role):
    def __init__(self, name: str):
        super().__init__(name)
        self._single_device: Device | str | None = None
        self._single_device_node: str | None = None
        """Actual Device object if it's local, str UUID of the device if remote, None if not set."""

    def on_set_engine(self, engine: IPCEngine):
        engine.events.add_listener(self.on_device_added, EngineEvents.DEVICE_ADDED)

    def add_device(self, device: Device):
        if self._single_device is None:
            super().add_device(device)
            self._single_device = device
            self._single_device_node = None
        else:
            raise ValueError("Only one device can be added to this role!")

    async def on_device_added(self, device_uuid: str, device_role: str, node_uuid: str):
        if device_role == self.name:
            if self._single_device is None:
                self._single_device = device_uuid
                self._single_device_node = node_uuid
            elif (isinstance(self._single_device, device.Device) and self._single_device.uuid != device_uuid) or \
                    (isinstance(self._single_device, str) and self._single_device != device_uuid):
                raise ValueError(
                    f"There is only supposed to be a single device for this role, but Device {device_uuid} was added "
                    f"when we already had one? Full list: {self.devices}"
                )

    async def on_incoming_chat(self, packet: connection.IPCPacket, node_uuid: str):
        logger.debug("Received incoming connection from %s %s", packet.origin_type, packet.origin_name)
        if self._single_device is None:
            logger.debug("Incoming connection denied due to no device being added.")
            conn = connection.IPCRoutedConnection.from_packet(self, packet, node_uuid)
            await conn.deny_chat_request(packet.origin_conn_uuid)
        else:
            logger.debug("Incoming connection redirected to single device.")
            device_uuid = self._single_device.uuid if isinstance(self._single_device, device.Device) else self._single_device
            dev_node_uuid = self.engine.uuid if self._single_device_node is None else self._single_device_node
            conn = connection.IPCRoutedConnection.from_packet(self, packet, node_uuid)
            await conn.redirect_chat_request(IPCClassType.DEVICE, device_uuid, dev_node_uuid, packet.origin_conn_uuid)




























